{-|
Module      : Gargantext.Core.Viz.Graph.Tools.IGraph
Description : Tools to build Graph
Copyright   : (c) CNRS, 2017-Present
License     : AGPL + CECILL v3
Maintainer  : team@gargantext.org
Stability   : experimental
Portability : POSIX

Reference:
* Gábor Csárdi, Tamás Nepusz: The igraph software package for complex network research. InterJournal Complex Systems, 1695, 2006.

-}

module Gargantext.Core.Viz.Graph.Tools.IGraph
  where

import Data.Serialize
import Graph.Types (ClusterNode(..))
import IGraph hiding (mkGraph, neighbors, edges, nodes, Node, Graph)
import Protolude
import Gargantext.Prelude (saveAsFileDebug)
import qualified Data.List                   as List
import qualified Data.Map.Strict             as Map
import qualified IGraph                      as IG
import qualified IGraph.Algorithms.Clique    as IG
import qualified IGraph.Algorithms.Community as IG
import qualified IGraph.Algorithms.Structure as IG
import qualified IGraph.Random               as IG
import qualified Data.Set                    as Set

------------------------------------------------------------------
-- | Main Types
type Graph_Undirected = IG.Graph 'U () ()
type Graph_Directed   = IG.Graph 'D () ()

type Node  = IG.Node
type Graph = IG.Graph

------------------------------------------------------------------
-- | Main Graph management Functions
neighbors :: IG.Graph d v e -> IG.Node -> [IG.Node]
neighbors = IG.neighbors

edges :: IG.Graph d v e -> [Edge]
edges = IG.edges

nodes :: IG.Graph d v e -> [IG.Node]
nodes = IG.nodes

------------------------------------------------------------------
-- | Partitions
maximalCliques :: IG.Graph d v e -> [[Int]]
maximalCliques g = IG.maximalCliques g (min',max')
  where
    min' = 0
    max' = 0

------------------------------------------------------------------
type Seed = Int

spinglass :: Seed -> Map (Int, Int) Double -> IO [ClusterNode]
spinglass s g = toClusterNode
             <$> List.concat
             <$> mapM (partitions_spinglass' s) g'
  where
    -- Non-connected graphs make spinglass crash
    g' = IG.decompose    -- decompose graph into connected components
       $ edgeList2UGraph -- convert into IGraph type
       $ Map.keys g      -- retrieve edges in the form of `(Int, Int)`

spinglass' :: Seed -> Map (Int, Int) Double -> IO [Set Int]
spinglass' s g = map Set.fromList
             <$> List.concat
             <$> mapM (partitions_spinglass' s) g'
  where
    -- Non-connected graphs make spinglass crash
    g' = IG.decompose    -- decompose graph into connected components
       $ edgeList2UGraph -- convert into IGraph type
       $ Map.keys g      -- retrieve edges in the form of `(Int, Int)`


-- | Tools to analyze graphs
partitions_spinglass' :: (Serialize v, Serialize e, Show v)
                         => Seed -> IG.Graph 'U v e -> IO [[v]]
partitions_spinglass' s g = do
  gen <- IG.withSeed s pure
  rawClusters <- IG.findCommunity g Nothing Nothing IG.spinglass gen
  let clusterLabels = (fmap . fmap) (IG.nodeLab g) rawClusters
  saveAsFileDebug "/tmp/res" clusterLabels
  pure clusterLabels


toClusterNode :: [[Int]] -> [ClusterNode]
toClusterNode ns = List.concat
                 $ map (\(cId, ns') -> map (\n -> ClusterNode n cId) ns')
                 $ List.zip [1..] ns

-- | Make an undirected IGraph graph from a list of edges between `Int`s.
-- The output graph's vertices are labeled with the original `Int`s, and the
-- edges are not labeled.
edgeList2UGraph :: [(Int, Int)] -> IG.Graph 'U Int ()
edgeList2UGraph edgeList =
 -- We're not using `IG.mkGraph` because of the following ticket:
 -- https://gitlab.iscpif.fr/gargantext/haskell-igraph/issues/4
  IG.fromLabeledEdges $ fmap (\edge -> (edge, ())) $ edgeList

mkGraphUfromEdges :: [(Int, Int)] -> Graph_Undirected
mkGraphUfromEdges es = IG.mkGraph (List.replicate n ()) $ zip es $ repeat ()
  where
    (a,b) = List.unzip es
    n = List.length (List.nub $ a <> b)
